package com.sillycycle.bagleyd.barrel;

/*
 * @(#)BarrelSolve.java
 *
 * Taken from James G. Nourse's The Simple Solutions to Cubic Puzzles.
 *
 * Copyright 2004 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>BarrelSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class BarrelSolve {
	private static JFrame frame;
	private static BarrelCanvas barrel;
	private static final boolean DEBUG = false;
	private static Random generator;
	//private static int counter; // debugging

	BarrelSolve(JFrame frame, BarrelCanvas barrel) {
		BarrelSolve.frame = frame;
		BarrelSolve.barrel = barrel;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	boolean moveATile(int direction, int tile, boolean shift) {
		if (!getFlag()) {
			return false;
		}
		return barrel.movePuzzle(direction, tile, shift, false,
			BarrelCanvas.DOUBLE);
	}

	static String colorName(int color) {
		switch (color) {
		case 0:
			return "Green";
		case 1:
			return "Red";
		case 2:
			return "Yellow";
		case 3:
			return "Orange";
		case 4:
			return "Blue";
		case 5:
			return "Black";
		default:
			return "None";
		}
	}

	static int nextInt(int n) {
		if (DEBUG) /* make it deterministic */
			return 1;
		return generator.nextInt(n);
	}

	static boolean checkLastColorMatch(int colors[], int last) {
		int i;

		for (i = 0; i < last; i++)
			if (colors[last] == colors[i])
				return true;
		return false;
	}

	static int ORIENTTILE(int x, int y) {
		return (barrel.tileOfPosition[y * barrel.tiles + x]);
	}
	static int COLORTILE(int x, int y) {
		return (ORIENTTILE(x, y) - 1) / (barrel.tiles - 2);
	}
	static int LEVELTILE(int x, int y) {
		return (ORIENTTILE(x, y) - 1) % (barrel.tiles - 2);
	}
	static int CENTER() {
		return COLORTILE(3, 2);
	}

	static int checkLevelSolved(int i) {
		int color, inc;

		if (barrel.orient) {
			for (int j = 0; j < barrel.faces; j++) {
				color = COLORTILE(1, j);
				inc = LEVELTILE(1, j);
				if (i >= 2 && i <= barrel.tiles - 2) {
					if (color != COLORTILE(i, j)) {
						return j;
					}
					if (inc + i != LEVELTILE(i, j)) {
						return j;
					}
				} else if (i == 1) {
					if (color == barrel.faces)
						return j;
					if (inc != 0)
						return j;
					if (j != 0 && color == COLORTILE(i, j))
						return j;
				} else if (i == barrel.faces &&
						(j & 1) == 1 &&
						color != barrel.faces &&
						LEVELTILE(barrel.tiles - 1, j) <
						barrel.faces) {
					return j;
				}
			}
		} else {
			for (int j = 0; j < barrel.faces; j++) {
				color = COLORTILE(1, j);
				if (i >= 2 && i < barrel.faces) {
					if (color != COLORTILE(i, j)) {
						return j;
					}
				} else if (i == 1) {
					if (color == barrel.faces)
						return j;
					for (int k = 0; k < j; k++)
						if (color == COLORTILE(i, k))
							return j;
				} else if (i == barrel.faces &&
						(j & 1) == 1 &&
						color != barrel.faces) {
					return j;
				}
			}
		}
		return barrel.faces;
	}

	static int findOtherColor(int level, int[] colors) {
		int i, j, color;
		int rand = nextInt(2);

		for (i = 0; i < barrel.faces; i++) {
			if (i == 0)
				j = (barrel.faces - 1) / 2;
			else if (rand == 0) {
				if (i % 2 == 0) {
					j = (barrel.faces - 1) / 2 + (i + 1) / 2;
				} else {
					j = (barrel.faces - 1) / 2 - (i + 1) / 2;
				}
			} else {
				if (i % 2 == 0) {
					j = (barrel.faces - 1) / 2 - (i + 1) / 2;
				} else {
					j = (barrel.faces - 1) / 2 + (i + 1) / 2;
				}
			}
			if ((j & 1) == 1 && (level == 2 || level == 5))
				continue;
			for (color = 0; color < barrel.faces; color++) {
				if (COLORTILE(level, j) == colors[color]) {
					break;
				}
				if (colors[color] == -1 ||
						color == barrel.faces - 1) {
					return j;
				}
			}
		}
		return barrel.faces;
	}

	static int findMatchColor(int level, int match) {
		int face;

		for (face = 0; face < barrel.faces; face++) {
			if (COLORTILE(level, face) == match) {
				return face;
			}
		}
		return barrel.faces;
	}

	static int findLevelOrient(int match) {
		int face, level;

		for (level = 1; level < barrel.tiles; level++) {
			for (face = 0; face < barrel.faces; face++) {
				if (ORIENTTILE(level, face) == match) {
					return level;
				}
			}
		}
		return barrel.tiles;
	}

	static int findFaceOrient(int level, int match) {
		int face;

		for (face = 0; face < barrel.faces; face++) {
			if (ORIENTTILE(level, face) == match) {
				return face;
			}
		}
		return barrel.faces;
	}

	void leftAway() {
		/* push top away */
		if (barrel.pairs) {
			moveATile(BarrelCanvas.TOP, 1, false);
		} else {
			moveATile(BarrelCanvas.TOP, 1, false);
			moveATile(BarrelCanvas.TOP, 2, false);
		}
	}

	void leftAway2() {
		leftAway();
		leftAway();
	}

	void leftTowards() {
		/* push top towards */
		if (barrel.pairs) {
			moveATile(BarrelCanvas.BOTTOM, 1, false);
		} else {
			moveATile(BarrelCanvas.BOTTOM, 1, false);
			moveATile(BarrelCanvas.BOTTOM, 2, false);
		}
	}

	void leftTowards2() {
		leftTowards();
		leftTowards();
	}

	void rightAway() {
		/* push top away */
		if (barrel.pairs) {
			moveATile(BarrelCanvas.TOP, barrel.tiles - 2, false);
		} else {
			moveATile(BarrelCanvas.TOP, barrel.tiles - 2, false);
			moveATile(BarrelCanvas.TOP, barrel.tiles - 3, false);
		}
	}

	void rightAway2() {
		rightAway();
		rightAway();
	}

	void rightTowards() {
		/* push top towards */
		if (barrel.pairs) {
			moveATile(BarrelCanvas.BOTTOM, barrel.tiles - 2, false);
		} else {
			moveATile(BarrelCanvas.BOTTOM, barrel.tiles - 2, false);
			moveATile(BarrelCanvas.BOTTOM, barrel.tiles - 3, false);
		}
	}

	void rightTowards2() {
		rightTowards();
		rightTowards();
	}

	void leftSlide() {
		moveATile(BarrelCanvas.LEFT, -1, false);
	}

	void rightSlide() {
		moveATile(BarrelCanvas.RIGHT, 1, false);
	}

	/*
	John Ewing & Czes Kosniowski, Puzzle it Out: "
	Cubes Groups and Puzzles, Cambridge University Press,
	New York, 1982, "Cylinder Puzzle", I only have pages 50-53
	*/

	/* Page 51 - 53 */
	void L() {
		leftTowards();leftSlide();leftAway();rightSlide();
	}
	void R() {
		rightTowards();leftSlide();rightAway();rightSlide();
	}
	void L_1() {
		leftSlide();leftTowards();rightSlide();leftAway();
	}
	void R_1() {
		leftSlide();rightTowards();rightSlide();rightAway();
	}
	/* Swaps 5 beads on rows 3 and 4 */
	void Lx7() {
		L();L();L();L();L();L();L();
	}
	void Rx7() {
		R();R();R();R();R();R();R();
	}
	void L_1x7() {
		L_1();L_1();L_1();L_1();L_1();L_1();L_1();
	}
	void R_1x7() {
		R_1();R_1();R_1();R_1();R_1();R_1();R_1();
	}
	/* commutatorx2 swap (9, 10, 11) the center beads */
	void commutator() {
		L();R_1();L_1();R();
	}
	void commutator_1() {
		R_1();L();R();L_1();
	}
	void commutatorx2() {
		commutator();commutator(); /* moves left */
	}
	void commutator_1x2() {
		commutator_1();commutator_1(); /* moves right */
	}
	void Lb() {
		leftAway();leftSlide();leftTowards();rightSlide();
	}
	void Rb() {
		rightAway();leftSlide();rightTowards();rightSlide();
	}
	void Lb_1() {
		leftSlide();leftAway();rightSlide();leftTowards();
	}
	void Rb_1() {
		leftSlide();rightAway();rightSlide();rightTowards();
	}
	/* Swaps 5 beads on rows 2 and 3 */
	void Lbx7() {
		Lb();Lb();Lb();Lb();Lb();Lb();Lb();
	}
	void Rbx7() {
		Rb();Rb();Rb();Rb();Rb();Rb();Rb();
	}
	void Lb_1x7() {
		Lb_1();Lb_1();Lb_1();Lb_1();Lb_1();Lb_1();Lb_1();
	}
	void Rb_1x7() {
		Rb_1();Rb_1();Rb_1();Rb_1();Rb_1();Rb_1();Rb_1();
	}
	void commutatorb() {
		Lb();Rb_1();Lb_1();Rb();
	}
	void commutatorb_1() {
		Rb_1();Lb();Rb();Lb_1();
	}
	/* commutatorx2 =~ commutatorbx2 */
	void commutatorbx2() {
		commutatorb();commutatorb(); /* moves left */
	}
	void commutator_1bx2() {
		commutatorb_1();commutatorb_1(); /* moves right */
	}
	/* There is at least one more page but do not have copy. */


	/*
	James G Nourse, The Simple Solutions to Cubic, Puzzles,
	Bantam Books, New York, November 1981, "The Barrel", pp 24-42.
	*/

	/* Page 27 */
	/* Near Swap 3rd column beads in 1 and 3 rows,
	   not for oriented T3230NS */
	/* void L2_11_3NS() {
		leftSlide();leftAway2();rightSlide();rightAway2();
		leftSlide();leftTowards2();rightSlide();rightTowards2();
	} */
	/* Near Swap 3rd column beads in 3 and 5 rows,
	   not for oriented T3234NS */
	/* void L2_11_19NS() {
		leftSlide();leftTowards2();rightSlide();rightTowards2();
		leftSlide();leftAway2();rightSlide();rightAway2();
	} */
	/* HS1 == R3230 */

	/* H2 Page 28 22->12 12->11, 5242 = 4232 */
	void L2_4232() {
		if (nextInt(2) == 1) {
			leftSlide();rightAway();rightSlide();rightTowards();
		} else {
			leftSlide();rightTowards();rightSlide();rightAway();
		}
	}
	void L2_5242() {
		L2_4232();
	}

	/* HS3 == HS2,HS2 T5232 */
	void L2_5232() {
		L2_5242();L2_4232();
	}

	/* HS4 21->11 */
	void L2_5032() {
		leftSlide();rightTowards();rightSlide();rightAway();
		leftSlide();rightTowards();rightSlide();rightTowards();
	}
	/* HS5 23->11 */
	void L2_5432() {
		leftSlide();rightAway();rightSlide();rightTowards();
		leftSlide();rightAway();rightSlide();rightAway();
	}

	/* First level is actually tricky,
	   short steps to line up for full insertion */
	/* Page 29 1B */
	void L1_B() {
		leftAway();rightAway();
		leftSlide();leftAway2();rightSlide();rightAway2();
		leftSlide();leftTowards2();rightSlide();
	}
	/* Page 30 1C */
	void L1_E() {
		rightAway();
		leftSlide();leftAway2();rightSlide();rightAway2();
		leftSlide();leftTowards2();rightSlide();
	}
	void L1_C() {
		leftTowards2();L1_E();
	}
	/* Page 30 1D */
	void L1_D() {
		leftTowards();
		leftSlide();leftAway();rightSlide();
	}
	void L1_F() {
		rightTowards();
		leftSlide();leftTowards2();rightSlide();rightTowards2();
		leftSlide();leftAway2();rightSlide();
	}
	void L1_G() {
		leftTowards();
		leftSlide();leftAway();rightSlide();
	}

	/* Page 33 R2B */
	/* HS1 Page 29 Add rightTowards2 cleaner */
	/*void L2_7_6() {
		leftSlide();leftAway2();rightSlide();rightAway2();
		leftSlide();leftTowards2();rightSlide();
	}*/
	/* Add rightAway2 cleaner */
	/*void L2_15_14() {
		leftSlide();leftTowards2();rightSlide();rightTowards2();
		leftSlide();leftAway2();rightSlide();
	}*/

	/* Page 30 R1C */
	/*void R19_6() {
		rightTowards2();R7_6();
	}*/
	/*void R7_14() {
		rightAway2();R15_14();
	}*/
	/* Page 31 R1F */
	/*void R7_18() {
		leftAway2();R7_6();leftTowards2();rightTowards2();
	}
	void R15_2() {
		leftTowards2();R15_14();leftAway2();rightAway2();
	}*/

	/* Page 31 */
	/* 7->6 */
	void L2_3121() {
		leftSlide();leftAway2();rightSlide();rightAway2();
		leftSlide();leftTowards2();rightSlide();
	}
	/* 6->19 */
	void L2_2134() {
		L2_3121();
	}
	/* 15->14 */
	void L2_3323() {
		leftSlide();leftTowards2();rightSlide();rightTowards2();
		leftSlide();leftAway2();rightSlide();
	}
	/* 14->3 */
	void L2_2330() {
		L2_3323();
	}
	/* 11->10 */
	void L2_3222() {
		if (nextInt(2) == 1) {
			leftTowards();rightTowards();L2_3323();leftAway();
		} else {
			leftAway();rightAway();L2_3121();leftTowards();
		}
	}
	/* 6->7 */
	void L2_2131() {
		leftSlide();leftAway2();rightSlide();rightTowards2();
		leftSlide();leftTowards2();rightSlide();
	}
	/* 14->15 */
	void L2_2333() {
		leftSlide();leftTowards2();rightSlide();rightAway2();
		leftSlide();leftAway2();rightSlide();
	}
	/* 10->11 */
	void L2_2232() {
		if (nextInt(2) == 1) {
			leftTowards();L2_2333();rightAway();leftAway();
		} else {
			leftAway();L2_2131();rightTowards();leftTowards();
		}
	}

	/* L2_11_6 == R1E */
	/* Cleaner but probably unnecessary, 4 changed, upside-down L */
	/*void L2_11_6C() {
		R11_6();rightAway2();
	}
	void L2_11_14C() {
		R11_14();rightTowards2();
	}*/


	/* Page 34, solve level 3, where 4 and 5 scratch area */
	/* 7->15 15->17 */
	void L3_3133() {
		if (nextInt(2) == 1) {
			leftSlide();rightTowards();rightSlide();rightAway2();
			leftSlide();rightAway2();rightSlide();rightTowards2();
			leftSlide();rightAway();rightSlide();rightAway2();
			leftSlide();rightAway2();rightSlide();rightTowards2();
		} else {
			leftSlide();rightAway();rightSlide();rightTowards2();
			leftSlide();rightTowards2();rightSlide();rightAway2();
			leftSlide();rightTowards();rightSlide();rightTowards2();
			leftSlide();rightTowards2();rightSlide();rightAway2();
		}
	}
	void L3_3331() {
		L3_3133();
	}

	/* Page 35 */
	/* 3->7 7->3 */
	void L3_3031() {
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightAway();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightAway();
	}
	void L3_3130() {
		L3_3031();
	}
	/* 15->19 19->15 */
	void L3_3334() {
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightTowards();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightTowards();
	}
	void L3_3433() {
		L3_3334();
	}

	/* 16->15 */
	void L3_4333() {
		leftSlide();rightTowards2();rightSlide();rightAway();
		leftSlide();rightTowards();rightSlide();rightAway2();
		leftSlide();rightTowards();rightSlide();rightAway();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
	}
	/* 8->7 */
	void L3_4131() {
		leftSlide();rightAway2();rightSlide();rightTowards();
		leftSlide();rightAway();rightSlide();rightTowards2();
		leftSlide();rightAway();rightSlide();rightTowards();
		leftSlide();rightAway2();rightSlide();rightAway2();
	}

	/* Page 35 */
	/* 12->15 */
	void L3_4233() {
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
	}
	/* 12->7 */
	void L3_4231() {
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightAway2();
	}

	/* 8->15 */
	void L3_4133() {
		leftSlide();rightAway();rightSlide();rightAway();
		leftSlide();rightAway();rightSlide();rightTowards2();
		leftSlide();rightAway();rightSlide();rightAway();
		leftSlide();rightAway();rightSlide();rightTowards2();
	}
	/* 16->7 */
	void L3_4331() {
		leftSlide();rightTowards();rightSlide();rightTowards();
		leftSlide();rightTowards();rightSlide();rightAway2();
		leftSlide();rightTowards();rightSlide();rightTowards();
		leftSlide();rightTowards();rightSlide();rightAway2();
	}

	/* 22->7 */
	void L3_5231() {
		leftSlide();rightAway();rightSlide();rightTowards2();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightTowards2();
		leftSlide();rightAway();rightSlide();rightAway2();
	}
	/* 22->15 */
	void L3_5233() {
		leftSlide();rightTowards();rightSlide();rightAway2();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightAway2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
	}

	/* 21->19 */
	void L3_5034() {
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightAway2();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightAway2();
	}
	/* 23->3 */
	void L3_5430() {
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightTowards2();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightTowards2();
	}

	/* Page 38, skip to solve the 3 plungers now, level 4 scratch area */
	/* 16->22 */
	void L5_4352() {
		leftSlide();rightAway2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightTowards();
		leftSlide();rightTowards2();rightSlide();rightTowards();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightAway2();rightSlide();rightTowards2();
	}
	/* 8->22 */
	void L5_4152() {
		leftSlide();rightTowards2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightAway();
		leftSlide();rightAway2();rightSlide();rightAway();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightTowards2();rightSlide();rightAway2();
	}

	/* 20->21 */
	void L5_4450() {
		leftSlide();rightTowards();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();
	}
	/* 4->23 */
	void L5_4054() {
		leftSlide();rightAway();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();
	}

	/* Page 40, just complete Level 4, All are swaps from this point */
	/* 12->8, 8->12 */
	void L4_4241() {
		leftSlide();rightTowards2();rightSlide();rightAway();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightAway();rightSlide();rightAway2();
		leftSlide();rightTowards();rightSlide();rightTowards();
		leftSlide();rightAway2();rightSlide();
	}

	/* 12->16, 16->12 */
	void L4_4243() {
		leftSlide();rightAway2();rightSlide();rightTowards();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightTowards();rightSlide();rightTowards2();
		leftSlide();rightAway();rightSlide();rightAway();
		leftSlide();rightTowards2();rightSlide();
	}

	/* 8->20, 20->8 */
	void L4_4144() {
		leftSlide();rightTowards();rightSlide();rightTowards();
		leftSlide();rightAway2();rightSlide();rightTowards();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightAway();
		leftSlide();rightTowards();rightSlide();rightTowards();
		leftSlide();rightAway2();rightSlide();rightTowards();
		leftSlide();rightTowards2();rightSlide();rightTowards2();
		leftSlide();rightTowards2();rightSlide();rightAway();
	}

	/* 16->4, 4->16 */
	void L4_4340() {
		leftSlide();rightAway();rightSlide();rightAway();
		leftSlide();rightTowards2();rightSlide();rightAway();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightTowards();
		leftSlide();rightAway();rightSlide();rightAway();
		leftSlide();rightTowards2();rightSlide();rightAway();
		leftSlide();rightAway2();rightSlide();rightAway2();
		leftSlide();rightAway2();rightSlide();rightTowards();
	}

	int findL1Color(int[] colors) {
		int searchLevel = 3;
		int piece = findOtherColor(searchLevel, colors);

		if (DEBUG) {
			System.out.println("1: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			rightTowards2();
			return CENTER();
		case 1:
			rightTowards();
			return CENTER();
		case 2:
			return CENTER();
		case 3:
			rightAway();
			return CENTER();
		case 4:
			rightAway2();
			return CENTER();
		default:
		}
		searchLevel = 4;
		piece = findOtherColor(searchLevel, colors);
		if (DEBUG) {
			System.out.println("1: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			rightTowards2();
			L2_4232();
			return CENTER();
		case 1:
			rightTowards();
			L2_4232();
			return CENTER();
		case 2:
			L2_4232();
			return CENTER();
		case 3:
			rightAway();
			L2_4232();
			return CENTER();
		case 4:
			rightAway2();
			L2_4232();
			return CENTER();
		default:
		}
		searchLevel = 5;
		piece = findOtherColor(searchLevel, colors);
		if (DEBUG) {
			System.out.println("1: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			L2_5032();
			return CENTER();
		case 2:
			L2_5232();
			return CENTER();
		case 4:
			L2_5432();
			return CENTER();
		default:
		}
		searchLevel = 2;
		piece = findOtherColor(searchLevel, colors);
		if (DEBUG) {
			System.out.println("1: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			leftTowards();
			L2_2131();
			leftAway();
			rightTowards();
			return CENTER();
		case 2:
			L2_2232();
			return CENTER();
		case 4:
			leftAway();
			L2_2333();
			leftTowards();
			rightAway();
			return CENTER();
		default:
		}
		System.out.println("findL1Color: Could not find!");
		return 5;
	}

	int findL2Color(int color) {
		int searchLevel = 3;
		int piece = findMatchColor(searchLevel, color);

		if (DEBUG) {
			System.out.println("2: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			rightTowards2();
			return CENTER();
		case 1:
			rightTowards();
			return CENTER();
		case 2:
			return CENTER();
		case 3:
			rightAway();
			return CENTER();
		case 4:
			rightAway2();
			return CENTER();
		default:
		}
		searchLevel = 4;
		piece = findMatchColor(searchLevel, color);
		if (DEBUG) {
			System.out.println("2: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			rightTowards2();
			L2_4232();
			return CENTER();
		case 1:
			rightTowards();
			L2_4232();
			return CENTER();
		case 2:
			L2_4232();
			return CENTER();
		case 3:
			rightAway();
			L2_4232();
			return CENTER();
		case 4:
			rightAway2();
			L2_4232();
			return CENTER();
		default:
		}
		searchLevel = 5;
		piece = findMatchColor(searchLevel, color);
		if (DEBUG) {
			System.out.println("2: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			L2_5032();
			return CENTER();
		case 2:
			L2_5232();
			return CENTER();
		case 4:
			L2_5432();
			return CENTER();
		default:
		}
		searchLevel = 2;
		piece = findMatchColor(searchLevel, color);
		if (DEBUG) {
			System.out.println("2: search level " +
				searchLevel + ", piece " + piece);
		}
		switch (piece) {
		case 0:
			leftTowards();
			L2_2131();
			leftAway();
			rightTowards();
			return CENTER();
		case 2:
			L2_2232();
			return CENTER();
		case 4:
			leftAway();
			L2_2333();
			leftTowards();
			rightAway();
			return CENTER();
		default:
		}
		System.out.println("findL2Color: Could not find!");
		return 5;
	}

	int findL3Color(int face, int color) {
		int searchLevel = 3;
		int piece = findMatchColor(searchLevel, color);

		if (DEBUG) {
			System.out.println("3: search level " +
				searchLevel + ", face " + face +
				", piece " + piece);
		}
		if (face == 2) {
			switch (piece) {
			case 0:
				rightTowards2();
				return searchLevel;
			case 1:
				rightTowards();
				return searchLevel;
			case 2:
				return searchLevel;
			case 3:
				rightAway();
				return searchLevel;
			case 4:
				rightAway2();
				return searchLevel;
			default:
			}
		} else {
			switch (piece) {
			case 0:
				switch (face) {
				case 1: /* 3->7 */
					L3_3031();
					return searchLevel;
				case 3: /* 3->15 */
					rightAway2();
					L3_3133();
					rightTowards2();
					return searchLevel;
				case 4: /* 3->19 */
					rightAway();
					L3_3334();
					rightTowards();
					return searchLevel;
				default:
					System.out.println("Should not get here");
				}
				return searchLevel;
			case 1:
				switch (face) {
				case 0: /* 7->3 */
					L3_3130();
					return searchLevel;
				case 3: /* 7->15 */
					L3_3133();
					return searchLevel;
				case 4: /* 7->19 */
					rightTowards2();
					L3_3133();
					rightAway2();
					return searchLevel;
				default:
					System.out.println("Should not get here");
				}
				return searchLevel;
			case 2:
				System.out.println("Should have been set first");
				return searchLevel;
	/* 7->15 15->17  L3_3133() move 2 */
	/* 3->7 7->3	 L3_3031() move 1 */
	/* 15->19 19->15 L3_3334() move 1 */
			case 3:
				switch (face) {
				case 0: /* 15->3 */
					rightAway2();
					L3_3133();
					rightTowards2();
					return searchLevel;
				case 1: /* 15->7 */
					L3_3331();
					return searchLevel;
				case 4: /* 15->19 */
					L3_3334();
					return searchLevel;
				default:
					System.out.println("Should not get here");
				}
				return searchLevel;
			case 4:
				switch (face) {
				case 0: /* 19->3 */
					rightTowards();
					L3_3031();
					rightAway();
					return searchLevel;
				case 1: /* 19->7 */
					rightTowards2();
					L3_3133();
					rightAway2();
					return searchLevel;
				case 3: /* 19->15 */
					L3_3433();
					return searchLevel;
				default:
					System.out.println("Should not get here");
				}
				return searchLevel;
			default:
			}
		}
		searchLevel = 4;
		piece = findMatchColor(searchLevel, color);
		if (DEBUG) {
			System.out.println("3: search level " +
				searchLevel + ", face " + face +
				", piece " + piece);
		}
		switch (piece) {
		case 0:
			switch (face) {
			case 0: /* 4->3 */
				rightTowards();
				L3_4131();
				rightAway();
				return searchLevel;
			case 1: /* 4->7 */
				rightTowards2();
				L3_4233();
				rightAway2();
				return searchLevel;
			case 2: /* 4->11 */
				rightTowards();
				L3_4133();
				rightAway();
				return searchLevel;
			case 3: /* 4->15 */
				rightAway2();
				L3_4331();
				rightTowards2();
				return searchLevel;
			case 4: /* 4->19 */
				rightTowards2();
				L3_4231();
				rightAway2();
				return searchLevel;
			default:
			}
			break;
		case 1:
			switch (face) {
			case 0: /* 8->3 */
				rightTowards();
				L3_4231();
				rightAway();
				return searchLevel;
			case 1: /* 8->7 */
				L3_4131();
				return searchLevel;
			case 2: /* 8->11 */
				rightTowards();
				L3_4233();
				rightAway();
				return searchLevel;
			case 3: /* 8->15 */
				L3_4133();
				return searchLevel;
			case 4: /* 8->19 */
				rightTowards2();
				L3_4331();
				rightAway2();
				return searchLevel;
			default:
			}
			break;
		case 2:
			switch (face) {
			case 0: /* 12->3 */
				rightTowards();
				L3_4331();
				rightAway();
				return searchLevel;
			case 1: /* 12->7 */
				L3_4231();
				return searchLevel;
			case 2: /* 12->11 */
				if (nextInt(2) == 1) {
					rightAway();
					L3_4131();
					rightTowards();
				} else {
					rightTowards();
					L3_4333();
					rightAway();
				}
				return searchLevel;
			case 3: /* 12->15 */
				L3_4233();
				return searchLevel;
			case 4: /* 12->19 */
				rightAway();
				L3_4133();
				rightTowards();
				return searchLevel;
			default:
			}
			break;
		case 3:
			switch (face) {
			case 0: /* 16->3 */
				rightAway2();
				L3_4133();
				rightTowards2();
				return searchLevel;
			case 1: /* 16->7 */
				L3_4331();
				return searchLevel;
			case 2: /* 16->11 */
				rightAway();
				L3_4231();
				rightTowards();
				return searchLevel;
			case 3: /* 16->15 */
				L3_4333();
				return searchLevel;
			case 4: /* 16->19 */
				rightAway();
				L3_4233();
				rightTowards();
				return searchLevel;
			default:
			}
			break;
	/* 16->15 L3_4333() */
	/* 8->7   L3_4131() */
	/* 12->15 L3_4233() */
	/* 12->7  L3_4231() */
	/* 8->15  L3_4133() */
	/* 16->7  L3_4331() */
		case 4:
			switch (face) {
			case 0: /* 20->3 */
				rightAway2();
				L3_4233();
				rightTowards2();
				return searchLevel;
			case 1: /* 20->7 */
				rightTowards2();
				L3_4133();
				rightAway2();
				return searchLevel;
			case 2: /* 20->11 */
				rightAway();
				L3_4331();
				rightTowards();
				return searchLevel;
			case 3: /* 20->15 */
				rightAway2();
				L3_4231();
				rightTowards2();
				return searchLevel;
			case 4: /* 20->19 */
				rightAway();
				L3_4333();
				rightTowards();
				return searchLevel;
			default:
			}
			break;
		default:
		}
		searchLevel = 5;
		if (DEBUG) {
			System.out.println("3: search level " +
				searchLevel + ", face " + face +
				", piece " + piece);
		}
		piece = findMatchColor(searchLevel, color);
		switch (piece) {
		case 0:
			switch (face) {
			case 0: /* 21->3 */
				rightAway();
				L3_5034();
				rightTowards();
				return searchLevel;
			case 1: /* 21->7 */
				rightAway2();
				L3_5034();
				rightTowards2();
				return searchLevel;
			case 2: /* 21->11 */
				rightTowards2();
				L3_5034();
				rightAway2();
				return searchLevel;
			case 3: /* 21->15 */
				rightTowards();
				L3_5034();
				rightAway();
				return searchLevel;
			case 4: /* 21->19 */
				L3_5034();
				return searchLevel;
			default:
			}
			break;
		case 2:
			switch (face) {
			case 0: /* 22->3 */
				rightTowards();
				L3_5231();
				rightAway();
				return searchLevel;
			case 1: /* 22->7 */
				L3_5231();
				return searchLevel;
			case 2: /* 22->11 */
				if (nextInt(2) == 1) {
					rightAway();
					L3_5231();
					rightTowards();
				} else {
					rightTowards();
					L3_5233();
					rightAway();
				}
				return searchLevel;
			case 3: /* 22->15 */
				L3_5233();
				return searchLevel;
			case 4: /* 22->19 */
				rightAway();
				L3_5233();
				rightTowards();
				return searchLevel;
			default:
			}
			break;
	/* 22->7  L3_5231() */
	/* 22->15 L3_5233() */
	/* 21->19 L3_5034() */
	/* 23->3  L3_5430() */
		case 4:
			switch (face) {
			case 0: /* 23->3 */
				L3_5430();
				return searchLevel;
			case 1: /* 23->7 */
				rightAway();
				L3_5430();
				rightTowards();
				return searchLevel;
			case 2: /* 23->11 */
				rightAway2();
				L3_5430();
				rightTowards2();
				return searchLevel;
			case 3: /* 23->15 */
				rightTowards2();
				L3_5430();
				rightAway2();
				return searchLevel;
			case 4: /* 23->19 */
				rightTowards();
				L3_5430();
				rightAway();
				return searchLevel;
			default:
			}
			break;
		default:
		}
		System.out.println("findL3Color: Could not find!");
		return 0;
	}

	boolean findL5Color(int face, int color) {
		int searchLevel = 4;
		int piece = findMatchColor(searchLevel, color);

		if (DEBUG) {
			System.out.println("5: search level " +
				searchLevel + ", face " + face +
				", piece " + piece);
		}
		switch (piece) {
		case 0:
			switch (face) {
			case 0: /* 4->21 */
				rightAway();
				L5_4450();
				rightTowards();
				return true;
			case 2: /* 4->22 */
				rightTowards();
				L5_4152();
				rightAway();
				return true;
			case 4: /* 4->23 */
				L5_4054();
				return true;
			default:
			}
			break;
		case 1:
			switch (face) {
			case 0: /* 8->21 */
				rightAway2();
				L5_4450();
				rightTowards2();
				return true;
			case 2: /* 8->22 */
				L5_4152();
				return true;
			case 4: /* 8->23 */
				rightAway();
				L5_4054();
				rightTowards();
				return true;
			default:
			}
			break;
		case 2:
			switch (face) {
			case 0: /* 12->21 */
				rightTowards2();
				L5_4450();
				rightAway2();
				return true;
			case 2: /* 12->22 */
				if (nextInt(2) == 1) {
					rightAway();
					L5_4152();
					rightTowards();
				} else {
					rightTowards();
					L5_4352();
					rightAway();
				}
				return true;
			case 4: /* 12->23 */
				rightAway2();
				L5_4054();
				rightTowards2();
				return true;
			default:
			}
			break;
	/* 16->22  L5_4352() */
	/* 8->22   L5_4152() */
	/* 20->21   L5_4450()*/
	/* 4->23   L5_4054() */
		case 3:
			switch (face) {
			case 0: /* 16->21 */
				rightTowards();
				L5_4450();
				rightAway();
				return true;
			case 2: /* 16->22 */
				L5_4352();
				return true;
			case 4: /* 16->23 */
				rightTowards2();
				L5_4054();
				rightAway2();
				return true;
			default:
			}
			break;
		case 4:
			switch (face) {
			case 0: /* 20->21 */
				L5_4450();
				return true;
			case 2: /* 20->22 */
				rightAway();
				L5_4352();
				rightTowards();
				return true;
			case 4: /* 20->23 */
				rightTowards();
				L5_4054();
				rightAway();
				return true;
			default:
			}
			break;
		default:
		}
		System.out.println("findL5Color: Could not find!");
		return false;
	}

	boolean findL4Color(int face, int color) {
		int searchLevel = 4;
		int piece = findMatchColor(searchLevel, color);

		if (DEBUG) {
			System.out.println("4: search level " +
				searchLevel + ", face " + face +
				", piece " + piece);
		}
		switch (piece) {
		case 0:
			switch (face) {
			case 1: /* 4->8 */
				rightTowards();
				L4_4241();
				rightAway();
				return true;
			case 2: /* 4->12 */
				rightAway2();
				L4_4340();
				rightTowards2();
				return true;
			case 3: /* 4->16 */
				L4_4340();
				return true;
			case 4: /* 4->20 */
				rightTowards2();
				L4_4241();
				rightAway2();
				return true;
			default:
			}
			break;
		case 1:
			switch (face) {
			case 0: /* 8->4 */
				rightTowards();
				L4_4241();
				rightAway();
				return true;
			case 2: /* 8->12 */
				L4_4241();
				return true;
			case 3: /* 8->16 */
				rightAway2();
				L4_4144();
				rightTowards2();
				return true;
			case 4: /* 8->20 */
				L4_4144();
				return true;
			default:
			}
			break;
		case 2:
			switch (face) {
			case 0: /* 12->4 */
				rightAway2();
				L4_4340();
				rightTowards2();
				return true;
			case 1: /* 12->8 */
				L4_4241();
				return true;
			case 3: /* 12->16 */
				L4_4243();
				return true;
			case 4: /* 12->20 */
				rightTowards2();
				L4_4144();
				rightAway2();
				return true;
			default:
			}
			break;
	/* 12->8   L4_4241() */
	/* 12->16  L4_4243() */
	/* 8->20   L4_4144() */
	/* 16->4   L4_4340() */
		case 3:
			switch (face) {
			case 0: /* 16->4 */
				L4_4340();
				return true;
			case 1: /* 16->8 */
				rightTowards2();
				L4_4340();
				rightAway2();
				return true;
			case 2: /* 16->12 */
				L4_4243();
				return true;
			case 4: /* 16->20 */
				rightAway();
				L4_4243();
				rightTowards();
				return true;
			default:
			}
			break;
		case 4:
			switch (face) {
			case 0: /* 20->4 */
				rightAway2();
				L4_4243();
				rightTowards2();
				return true;
			case 1: /* 20->8 */
				L4_4144();
				return true;
			case 2: /* 20->12 */
				rightTowards2();
				L4_4144();
				rightAway2();
				return true;
			case 3: /* 20->16 */
				rightAway();
				L4_4243();
				rightTowards();
				return true;
			default:
			}
			break;
		default:
		}
		System.out.println("findL4Color: Could not find!");
		return false;
	}

	boolean solveRightColor(int level, int face) {
		int piece, check, searchLevel = 4;

		if (level == 5)
			piece = barrel.faces;
		else
			piece = COLORTILE(1, face);
		check = COLORTILE(level, face);
		if (piece == check)
			return true;
		switch (level) {
		case 3:
			searchLevel = findL3Color(face, piece);
			break;
		case 4:
			findL4Color(face, piece);
			break;
		default:
			findL5Color(face, piece);
			break;
		}
		check = COLORTILE(level, face);
		if (piece != check) {
			System.out.println("error level " + searchLevel +
				", face " + face + ", piece " + piece +
				", check " + check);
			return false;
		}
		return true;
	}

	void solveL0() {
		if (barrel.tileOfPosition[0] > 0)
			rightSlide();
		if (barrel.spacePosition[2] == 0) {
			moveATile(BarrelCanvas.TOP, 0, true);
		} else if (barrel.spacePosition[2] == 6) {
			moveATile(BarrelCanvas.TOP, 0, true);
			moveATile(BarrelCanvas.TOP, 0, true);
		} else if (barrel.spacePosition[1] == barrel.tiles) {
			moveATile(BarrelCanvas.BOTTOM, 0, true);
		} else if (barrel.spacePosition[1] == 0) {
			moveATile(BarrelCanvas.BOTTOM, 0, true);
			moveATile(BarrelCanvas.BOTTOM, 0, true);
		}
	}

	int moveToCenter(int level, int face) {
		if (DEBUG) {
			System.out.println("moveToCenter level " + level +
				", face " + face);
		}
		/* move to 11 */
		switch (level) {
		case 1:
			switch (face) {
			case 0:
				leftTowards2();
				Lbx7();
				leftAway2();
				return CENTER();
			case 1:
				leftTowards();
				L_1x7();
				leftAway();
				return moveToCenter(2, 0);
			case 3:
				leftAway();
				Lb_1x7();
				leftTowards();
				return moveToCenter(2, 4);
			case 4:
				leftAway2();
				Lx7();
				leftTowards2();
				return CENTER();
			default:
			}
			break;
		case 2:
			switch (face) {
			case 0:
				leftTowards();
				rightAway();
				L2_2131();
				rightTowards();
				leftAway();
				return CENTER();
			case 1:
				rightAway();
				L2_2131();
				rightTowards();
				return CENTER();
			case 2:
				L2_2232();
				return CENTER();
			case 3:
				rightTowards();
				L2_2333();
				rightAway();
				return CENTER();
			case 4:
				leftAway();
				rightTowards();
				L2_2333();
				rightAway();
				leftTowards();
				return CENTER();
			default:
			}
			break;
		case 3:
			switch (face) {
			case 0:
				rightTowards2();
				return CENTER();
			case 1:
				rightTowards();
				return CENTER();
			case 2:
				return CENTER();
			case 3:
				rightAway();
				return CENTER();
			case 4:
				rightAway2();
				return CENTER();
			default:
			}
			break;
		case 4:
			switch (face) {
			case 0: /* 4->11 */
				rightTowards();
				L3_4133();
				rightAway();
				return CENTER();
			case 1: /* 8->11 */
				rightTowards();
				L3_4233();
				rightAway();
				return CENTER();
			case 2: /* 12->11 */
				if (nextInt(2) == 1) {
					rightAway();
					L3_4131();
					rightTowards();
				} else {
					rightTowards();
					L3_4333();
					rightAway();
				}
				return CENTER();
			case 3: /* 16->11 */
				rightAway();
				L3_4231();
				rightTowards();
				return CENTER();
			case 4: /* 20->11 */
				rightAway();
				L3_4331();
				rightTowards();
				return CENTER();
			default:
			}
			break;
		case 5:
			switch (face) {
			case 0: /* 21->11 */
				rightTowards2();
				L3_5034();
				rightAway2();
				return CENTER();
			case 2: /* 22->11 */
				if (nextInt(2) == 1) {
					rightAway();
					L3_5231();
					rightTowards();
				} else {
					rightTowards();
					L3_5233();
					rightAway();
				}
				return CENTER();
			case 4: /* 23->11 */
				rightAway2();
				L3_5430();
				rightTowards2();
				return CENTER();
			default:
			}
			break;
		default:
			break;
		}
		return -1;
	}

	boolean solveL1Orient() {
		int[] orient = {-1, -1, -1, -1, -1};
		int piece = checkLevelSolved(1);
		int currentLevel, currentFace;
		int ok;

		if (piece == barrel.faces) {
			return true;
		}
		piece = 1;
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		/* G1 B17 Y9 R5 O13 */
		if (ORIENTTILE(1, 2) != piece) {
			currentLevel = findLevelOrient(piece);
			currentFace = findFaceOrient(currentLevel, piece);
			if (currentLevel == 1) {
				switch (currentFace) {
				case 0:
					leftTowards2();
					break;
				case 1:
					leftTowards();
					break;
				case 3:
					leftAway();
					break;
				case 4:
					leftAway2();
					break;
				default:
				}
			} else {
				if (currentLevel == 2) {
					switch (currentFace) {
					case 0:
						leftTowards2();
						break;
					case 1:
						leftTowards();
						break;
					case 3:
						leftAway();
						break;
					case 4:
						leftAway2();
						break;
					default:
					}
					leftSlide();
					if (nextInt(2) == 1) {
						leftAway();
						rightSlide();
						leftTowards();
					} else {
						leftTowards();
						rightSlide();
						leftAway();
					}
				} else {
					moveToCenter(currentLevel,
						currentFace);
					leftSlide();
					if (nextInt(2) == 1) {
						leftAway();
						rightSlide();
						leftTowards();
					} else {
						leftTowards();
						rightSlide();
						leftAway();
					}
					leftSlide();
					if (nextInt(2) == 1) {
						leftAway();
						rightSlide();
						leftTowards();
					} else {
						leftTowards();
						rightSlide();
						leftAway();
					}
				}
			}
		}
		piece = 17;
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		if (ORIENTTILE(2, 2) != piece) {
			currentLevel = findLevelOrient(piece);
			currentFace = findFaceOrient(currentLevel, piece);
			ok = moveToCenter(currentLevel, currentFace);
			L2_3222();
			if (DEBUG) {
				System.out.println("solving level 1, " +
					currentLevel + " " +
					currentFace + " " + ok);
			}
		}
		orient[0] = ORIENTTILE(1, 2);
		orient[1] = ORIENTTILE(2, 2);
		if (DEBUG) {
			System.out.println(
				"1A: 1,2: " + orient[0] +
				", 2,2: " + orient[1]);
		}
		piece = 9;
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		if (ORIENTTILE(3, 2) != piece) {
			currentLevel = findLevelOrient(piece);
			currentFace = findFaceOrient(currentLevel, piece);
			ok = moveToCenter(currentLevel, currentFace);
			if (DEBUG) {
				System.out.println("solving level 1, " +
					currentLevel + " " +
					currentFace + " " + ok);
			}
		}
		orient[0] = ORIENTTILE(1, 2);
		orient[1] = ORIENTTILE(2, 2);
		orient[2] = ORIENTTILE(3, 2);
		if (DEBUG) {
			System.out.println(
				"1B: 1,2: " + orient[0] +
				", 2,2: " + orient[1] +
				", 3,2: " + orient[2]);
		}
		orient[0] = ORIENTTILE(1, 3);
		orient[1] = ORIENTTILE(2, 3);
		orient[2] = ORIENTTILE(2, 1);
		if (DEBUG) {
			System.out.println(
				"1C: 1,3: " + orient[0] +
				", 2,3: " + orient[1] +
				", 2,1: " + orient[2]);
		}
		orient[0] = ORIENTTILE(1, 4);
		orient[1] = ORIENTTILE(1, 3);
		orient[2] = ORIENTTILE(1, 1);
		orient[3] = ORIENTTILE(2, 1); /* see if already in place */
		if (DEBUG) {
			System.out.println(
				"1E: 1,4: " + orient[0] +
				", 1,3: " + orient[1] +
				", 1,1: " + orient[2] +
				", 2,1: " + orient[3]);
		}
		piece = 5;
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		if (ORIENTTILE(2, 1) != piece) {
			currentLevel = findLevelOrient(piece);
			currentFace = findFaceOrient(currentLevel, piece);
			ok = moveToCenter(currentLevel, currentFace);
			if (DEBUG) {
				System.out.println("solving level 1, " +
					currentLevel + " " +
					currentFace + " " + ok);
			}
			L1_E();
		}
		orient[0] = ORIENTTILE(1, 4);
		orient[1] = ORIENTTILE(1, 3);
		orient[2] = ORIENTTILE(1, 1);
		orient[3] = ORIENTTILE(2, 1);
		orient[4] = ORIENTTILE(2, 3); /* see if already in place */
		if (DEBUG) {
			System.out.println(
				"1F: 1,4: " + orient[0] +
				", 1,3: " + orient[1] +
				", 1,1: " + orient[2] +
				", 2,1: " + orient[3] +
				", 2,3: " + orient[4]);
		}
		piece = 13;
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		if (ORIENTTILE(2, 3) != piece) {
			currentLevel = findLevelOrient(piece);
			currentFace = findFaceOrient(currentLevel, piece);
			ok = moveToCenter(currentLevel, currentFace);
			if (DEBUG) {
				System.out.println("solving level 1, " +
					currentLevel + " " +
					currentFace + " " + ok);
			}
			L1_F();
		}
		orient[0] = ORIENTTILE(1, 4);
		orient[1] = ORIENTTILE(1, 3);
		orient[2] = ORIENTTILE(1, 1);
		orient[3] = ORIENTTILE(2, 1);
		orient[4] = ORIENTTILE(2, 3); /* see if already in place */
		if (DEBUG) {
			System.out.println(
				"1G: 1,4: " + orient[0] +
				", 1,3: " + orient[1] +
				", 1,1: " + orient[2] +
				", 2,1: " + orient[3] +
				", 2,3: " + orient[4]);
		}
		L1_G();
		orient[0] = ORIENTTILE(1, 0);
		orient[1] = ORIENTTILE(1, 4);
		orient[2] = ORIENTTILE(1, 2);
		orient[3] = ORIENTTILE(1, 1);
		orient[4] = ORIENTTILE(1, 3); /* see if already in place */
		if (DEBUG) {
			System.out.println(
				"1H: 1,0: " + orient[0] +
				", 1,4: " + orient[1] +
				", 1,2: " + orient[2] +
				", 1,1: " + orient[3] +
				", 1,3: " + orient[4]);
		}
		return true;
	}

	boolean solveL1Color() {
		int[] colors = {-1, -1, -1, -1, -1, -1};
		int piece = checkLevelSolved(1);

		if (piece == barrel.faces) {
			return true;
		}
		if (DEBUG) {
			System.out.println("solving level 1, piece " + piece);
		}
		/* rotate first column to non black, put in 9 */
		if (COLORTILE(1, 2) == barrel.faces) {
			if (nextInt(2) == 1) {
				if (COLORTILE(1, 1) != barrel.faces) {
					leftTowards();
				} else if (COLORTILE(1, 3) ==
						barrel.faces) {
					leftTowards2();
				} else {
					leftAway();
				}
			} else {
				if (COLORTILE(1, 3) != barrel.faces) {
					leftAway();
				} else if (COLORTILE(1, 1) ==
						barrel.faces) {
					leftAway2();
				} else {
					leftTowards();
				}
			}
		}
		colors[0] = barrel.faces;
		colors[1] = COLORTILE(1, 2);
		colors[2] = COLORTILE(2, 2);
		if (DEBUG) {
			System.out.println(
				"1A: 1,2: " + colorName(colors[1]) +
				", 2,2: " + colorName(colors[2]));
		}
		{
			if (checkLastColorMatch(colors, 2)) {
				colors[2] = -1;
				/* rotate 3rd column to non black and non matching */
				colors[2] = findL1Color(colors);
				L1_B();
			} else {
				leftAway();
			}
			colors[3] = COLORTILE(2, 4); /* see if already in place */
			if (DEBUG) {
				System.out.println(
					"1B: 1,1: " + colorName(colors[1]) +
					", 2,1: " + colorName(colors[2]) +
					", 2,4: " + colorName(colors[3]));
			}
			leftTowards2();
			if (checkLastColorMatch(colors, 3)) {
				colors[3] = -1;
				/* rotate 3rd column to non black and non matching */
				colors[3] = findL1Color(colors);
				L1_E();
			}
			if (DEBUG) {
				System.out.println(
					"1C: 1,3: " + colorName(colors[1]) +
					", 2,3: " + colorName(colors[2]) +
					", 2,1: " + colorName(colors[3]));
			}
			L1_D();
			colors[4] = COLORTILE(2, 1); /* see if already in place */
			if (DEBUG) {
				System.out.println(
					"1E: 1,4: " + colorName(colors[1]) +
					", 1,3: " + colorName(colors[2]) +
					", 1,1: " + colorName(colors[3]) +
					", 2,1: " + colorName(colors[4]));
			}
			if (checkLastColorMatch(colors, 4)) {
				colors[4] = -1;
				/* rotate 3rd column to non black and non matching */
				colors[4] = findL1Color(colors);
				L1_E();
			}
			colors[5] = COLORTILE(2, 3); /* see if already in place */
			if (DEBUG) {
				System.out.println(
					"1F: 1,4: " + colorName(colors[1]) +
					", 1,3: " + colorName(colors[2]) +
					", 1,1: " + colorName(colors[3]) +
					", 2,1: " + colorName(colors[4]) +
					", 2,3: " + colorName(colors[5]));
			}
			if (checkLastColorMatch(colors, 5)) {
				colors[5] = -1;
				/* rotate 3rd column to non black and non matching */
				colors[5] = findL1Color(colors);
				L1_F();
			}
			if (DEBUG) {
				System.out.println(
					"1G: 1,4: " + colorName(colors[1]) +
					", 1,3: " + colorName(colors[2]) +
					", 1,1: " + colorName(colors[3]) +
					", 2,1: " + colorName(colors[4]) +
					", 2,3: " + colorName(colors[5]));
			}
			L1_G();
		}
		piece = checkLevelSolved(1);
		if (piece != barrel.faces) {
			System.out.println("Error in level 1, aborting, " +
				piece);
			return false;
		}
		return true;
	}

	boolean solveL2Color() {
		/*int faces[MAX_FACES] = {1, 3, 2, 0, 5};*/
		int piece = checkLevelSolved(2);

		if (piece == barrel.faces) {
			return true;
		}
		if (DEBUG) {
			System.out.println("solving level 2, piece " + piece);
		}
		{
			piece = COLORTILE(1, 1);
			if (piece != COLORTILE(2, 1)) {
				/* rotate 3rd column to non black and non matching */
				piece = findL2Color(piece);
				if (piece == barrel.faces)
					System.out.println("error in level 2");
				leftTowards();
				L2_3222();
				leftAway();
			}
			piece = COLORTILE(1, 3);
			if (piece != COLORTILE(2, 3)) {
				/* rotate 3rd column to non black and non matching */
				piece = findL2Color(piece);
				if (piece == barrel.faces)
					System.out.println("error in level 2");
				leftAway();
				L2_3222();
				leftTowards();
			}
			piece = COLORTILE(1, 2);
			if (piece != COLORTILE(2, 2)) {
				/* rotate 3rd column to non black and non matching */
				piece = findL2Color(piece);
				if (piece == barrel.faces)
					System.out.println("error in level 2");
				L2_3222();
			}
			piece = COLORTILE(1, 0);
			if (piece != COLORTILE(2, 0)) {
				/* rotate 3rd column to non black and non matching */
				piece = findL2Color(piece);
				if (piece == barrel.faces)
					System.out.println("error in level 2");
				leftTowards2();
				L2_3222();
				leftAway2();
			}
			piece = COLORTILE(1, 4);
			if (piece != COLORTILE(2, 4)) {
				/* rotate 3rd column to non black and non matching */
				piece = findL2Color(piece);
				if (piece == barrel.faces)
					System.out.println("error in level 2");
				leftAway2();
				L2_3222();
				leftTowards2();
			}
		}
		return true;
	}

	boolean solveL3Color() {
		int piece = checkLevelSolved(3);

		if (piece == barrel.faces) {
			return true;
		}
		if (DEBUG) {
			System.out.println("solving level 3, piece " + piece);
		}
		if (!solveRightColor(3, 2))
			return false;
		if (nextInt(2) == 1) {
			if (!solveRightColor(3, 1))
				return false;
			if (!solveRightColor(3, 3))
				return false;
		} else {
			if (!solveRightColor(3, 3))
				return false;
			if (!solveRightColor(3, 1))
				return false;
		}
		if (nextInt(2) == 1) {
			if (!solveRightColor(3, 0))
				return false;
			if (!solveRightColor(3, 4))
				return false;
		} else {
			if (!solveRightColor(3, 4))
				return false;
			if (!solveRightColor(3, 0))
				return false;
		}
		return true;
	}

	boolean solveL5Color() {
		int piece = checkLevelSolved(5);

		if (piece == barrel.faces) {
			return true;
		}
		if (DEBUG) {
			System.out.println("solving level 5, piece " + piece);
		}
		if (!solveRightColor(5, 2))
			return false;
		if (nextInt(2) == 1) {
			if (!solveRightColor(5, 0))
				return false;
			if (!solveRightColor(5, 4))
				return false;
		} else {
			if (!solveRightColor(5, 4))
				return false;
			if (!solveRightColor(5, 0))
				return false;
		}
		return true;
	}

	boolean solveL4Color() {
		int piece = checkLevelSolved(4);

		if (piece == barrel.faces) {
			return true;
		}
		if (DEBUG) {
			System.out.println("solving level 4, piece " + piece);
		}
		if (!solveRightColor(4, 2))
			return false;
		if (nextInt(2) == 1) {
			if (!solveRightColor(4, 1))
				return false;
			if (!solveRightColor(4, 3))
				return false;
		} else {
			if (!solveRightColor(4, 3))
				return false;
			if (!solveRightColor(4, 1))
				return false;
		}
		if (nextInt(2) == 1) {
			if (!solveRightColor(4, 0))
				return false;
			if (!solveRightColor(4, 4))
				return false;
		} else {
			if (!solveRightColor(4, 4))
				return false;
			if (!solveRightColor(4, 0))
				return false;
		}
		return true;
	}

	/* This procedure coordinates the solution process. */
	void solveSomeTiles() {
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_RESET);
		if (!barrel.checkSolved()) {
			solveL0();
			if (barrel.orient) {
				if (!solveL1Orient())
					return;
			} else {
				if (!solveL1Color())
					return;
				if (!solveL2Color())
					return;
				if (!solveL3Color())
					return;
				if (!solveL5Color())
					return;
				if (!solveL4Color())
					return;
			}
		}
		((BarrelFrame) frame).callback(
			BarrelInterface.ACTION_COMPUTED);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}
		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solveSomeTiles();
					runFlag = false;
				}
			}
		}
	}
	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
