package com.sillycycle.bagleyd.barrel;

/*
 * @(#)BarrelStack.java
 *
 * Copyright 2004 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Deque;
import java.util.LinkedList;

/**
 * The <code>BarrelStack</code> class allows one to undo and redo
 * in the <code>BarrelFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class BarrelStack {
	//int direction, shift;
	private int packed; /* This makes assumptions on the data. */
	private int tile;   /* Do not make assumptions on this one. */
	private int count;
	private final Deque<String> stack = new LinkedList<>();

	private void writeMoveDirection(int direction) {
		//this.direction = direction;
		packed += ((direction & 0xF) << 4);
	}

	private void writemoveShift(int shift) {
		//this.shift = shift;
		packed += shift & 0xF;
	}

	private void writemoveTile(int aTile) {
		this.tile = aTile;
	}

	int readMoveDirection() {
		//return this.direction;
		return ((packed >> 4) & 0xF);
	}

	int readmoveShift() {
		//return this.shift;
		return (packed & 0xF);
	}

	int readmoveTile() {
		return this.tile;
	}

	void setMove(int direction, int tile, int shift) {
		packed = 0;
		writeMoveDirection(direction);
		writemoveShift(shift);
		writemoveTile(tile);
		stack.push(Integer.toString(packed));
		stack.push(Integer.toString(tile));
		count++;
	}

	void getMove() {
		tile = Integer.parseInt(stack.pop());
		packed = Integer.parseInt(stack.pop());
		count--;
	}

	boolean madeMoves() {
		return !stack.isEmpty();
	}

	void flushMoves() {
		while (madeMoves())
			getMove();
		count = 0; // probably do not need
	}

	int numMoves() {
		return count;
	}
}
